from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.events.base import PyLavEvent
from pylav.nodes.api.responses.plugins import SegmentSkipped, SegmentsLoaded

__all__ = ("SegmentSkippedEvent", "SegmentsLoadedEvent")

if TYPE_CHECKING:
    from pylav.nodes.node import Node
    from pylav.players.player import Player


class SegmentSkippedEvent(PyLavEvent):
    """This event is dispatched when a segment is skipped."""

    __slots__ = ()

    def __init__(self, player: Player, node: Node, event_object: SegmentSkipped) -> None:
        self.player = player
        self.segment = event_object.segment
        self.node = node
        self.event = event_object


class SegmentsLoadedEvent(PyLavEvent):
    """This event is dispatched when segments are loaded."""

    __slots__ = ()

    def __init__(self, player: Player, node: Node, event_object: SegmentsLoaded) -> None:
        self.player = player
        self.segments = event_object.segments
        self.node = node
        self.event = event_object



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.events.base import PyLavEvent
from pylav.nodes.api.responses.websocket import TrackStart

if TYPE_CHECKING:
    from pylav.nodes.node import Node
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class TrackStartEvent(PyLavEvent):
    """This event is dispatched when the player starts to play a track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        self.player = player
        self.track = track
        self.node = node
        self.event = event_object


class TrackStartYouTubeEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a YouTube track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


# noinspection SpellCheckingInspection
class TrackStartClypitEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Clyp.it track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartGetYarnEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a GetYarn track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartMixCloudEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a MixCloud track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartOCRMixEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a OCR Mix track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartPornHubEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Pornhub track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartRedditEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Reddit track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


# noinspection SpellCheckingInspection
class TrackStartSoundgasmEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Soundgasm track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartTikTokEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a TikTok track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartSpotifyEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Spotify track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartDeezerEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Deezer track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartYandexMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Yandex Music track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartAppleMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play an Apple Music track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartBandcampEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Bandcamp track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartYouTubeMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a YouTube Music track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartSoundCloudEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a SoundCloud track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartTwitchEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Twitch track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartHTTPEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play an HTTP track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartLocalFileEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a local file track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartNicoNicoEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a NicoNico track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartVimeoEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Vimeo track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartSpeakEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Speak track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartGCTTSEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Google Cloud TTS track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

from pylav.logging import getLogger
from pylav.storage.database.tables.player_state import PlayerStateRow
from pylav.storage.models.player.state import PlayerState
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

if TYPE_CHECKING:
    from pylav.core.client import Client
LOGGER = getLogger("PyLav.Database.Controller.Player.State")


class PlayerStateController:
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    async def save_players(self, players: list[JSON_DICT_TYPE]) -> None:
        for player in players:
            await self.save_player(player)
        LOGGER.debug("Saved %s players", len(players))

    async def save_player(self, player: JSON_DICT_TYPE) -> None:
        await PlayerState(bot=self.client.bot.user.id, **player).save()
        LOGGER.trace("Saved player %s", player.get("id"))

    async def fetch_player(self, guild_id: int) -> PlayerState | None:
        return await PlayerState.get(bot_id=self._client.bot.user.id, guild_id=guild_id)

    async def fetch_all_players(self) -> AsyncIterator[PlayerState]:
        for entry in await PlayerStateRow.select(
            *(PlayerStateRow.all_columns(exclude=[PlayerStateRow.primary_key]))
        ).where(PlayerStateRow.bot == self.client.bot.user.id):
            yield PlayerState(**entry)

    async def delete_player(self, guild_id: int) -> None:
        await PlayerStateRow.delete().where(
            (PlayerStateRow.bot == self.client.bot.user.id) & (PlayerStateRow.id == guild_id)
        )

    async def delete_all_players(self) -> None:
        await PlayerStateRow.delete().where(PlayerStateRow.bot == self.client.bot.user.id)



from __future__ import annotations

from piccolo.columns import M2M, LazyTableReference, Text, Timestamptz
from piccolo.columns.defaults.timestamptz import TimestamptzNow
from piccolo.table import Table

from pylav.helpers.time import get_now_utc
from pylav.storage.database.tables.misc import DATABASE_ENGINE


class QueryRow(Table, db=DATABASE_ENGINE, tablename="query"):
    identifier = Text(null=False, index=True, primary_key=True)
    name = Text(null=True, default=None)
    last_updated = Timestamptz(null=False, index=True, default=TimestamptzNow(), auto_update=get_now_utc)
    tracks = M2M(LazyTableReference("TrackToQueries", module_path="pylav.storage.database.tables.m2m"))



from __future__ import annotations

from piccolo.columns import JSONB, Array, BigInt, Boolean, Integer, Text
from piccolo.table import Table

from pylav.storage.database.tables.misc import DATABASE_ENGINE


class NodeRow(Table, db=DATABASE_ENGINE, tablename="node"):
    id = BigInt(primary_key=True, index=True)
    name = Text(null=False)
    ssl = Boolean(null=False, default=False)
    resume_key = Text(null=True, default=None)
    resume_timeout = Integer(null=False, default=600)
    reconnect_attempts = Integer(null=False, default=-1)
    search_only = Boolean(null=False, default=False)
    managed = Boolean(null=False, default=False)
    disabled_sources = Array(null=False, default=[], base_column=Text())
    extras = JSONB(null=True, default={})
    yaml = JSONB(null=True, default={})



from __future__ import annotations

import dataclasses
from typing import Annotated

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Echo:
    delay: Annotated[int | None, ValueRange(min=0, max=float("inf"))] = None
    decay: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None

    def to_dict(self) -> dict[str, float | int]:
        response: dict[str, float | int] = {}
        if self.delay is not None:
            response["delay"] = self.delay
        if self.decay is not None:
            response["decay"] = self.decay
        return response



from __future__ import annotations

import contextlib
import socket

import aiohttp
from cashews import Cache  # type: ignore
from yarl import URL

from pylav.compat import json
from pylav.exceptions.base import PyLavException
from pylav.logging import getLogger

LOGGER = getLogger("PyLav.extension.RadioBrowser")


CACHE = Cache("RADIO")
CACHE.setup("mem://?check_interval=10", size=1_000_000, enable=True)


class Error(PyLavException):
    """Base class for all exceptions raised by this module"""


class RDNSLookupError(Error):
    __slots__ = ("ip", "port")

    def __init__(self, ip: str) -> None:
        self.ip = ip
        self.error_msg = f"There was a problem with performing reverse dns lookup for ip: {ip}"
        super().__init__(self.error_msg)


@CACHE(ttl=3600, prefix="fetch_servers")
async def fetch_servers() -> set[str]:
    """
    Get IP of all currently available `Radio Browser` servers.
    Returns:
        set: List of addresses.
    """
    try:
        async with aiohttp.ClientSession(json_serialize=json.dumps) as session:
            async with session.get("http://all.api.radio-browser.info/json/servers") as response:
                data = await response.json(loads=json.loads)
    except socket.gaierror:
        return set()
    else:
        return {server["name"] for server in data}


@CACHE(ttl=300, prefix="pick_base_url")
async def pick_base_url(session: aiohttp.ClientSession) -> URL | None:
    servers = await fetch_servers()
    if not servers:
        LOGGER.warning("RadioBrowser API seems to be down at the moment, disabling Radio functionality.")
        return None
    for host in servers:
        with contextlib.suppress(Exception):
            async with session.get(f"https://{host}/json/stats") as response:
                if response.status == 200:
                    return URL(f"https://{host}")
                LOGGER.verbose("Error interacting with %s: %s", host, response.status)
    LOGGER.error("All the following hosts for the RadioBrowser API are broken: %s", ", ".join(servers))



from __future__ import annotations

import discord

MINUS = discord.PartialEmoji(name="minus", animated=False, id=965672202013925447)
PLAYLIST = discord.PartialEmoji(name="playlist", animated=False, id=965672202093621319)
LOOP = discord.PartialEmoji(name="loop", animated=False, id=965672202143928362)
FOLDER = discord.PartialEmoji(name="folder", animated=False, id=965672202395586601)
POWER = discord.PartialEmoji(name="power", animated=False, id=965672202395586691)
SETTINGS = discord.PartialEmoji(name="settings", animated=False, id=965672202395590696)
# noinspection SpellCheckingInspection
VOLUME_DOWN = discord.PartialEmoji(name="volumedown", animated=False, id=965672202399801374)
REPEAT = discord.PartialEmoji(name="repeat", animated=False, id=965672202412388352)
PLUS = discord.PartialEmoji(name="plus", animated=False, id=965672202416570368)
NEXT = discord.PartialEmoji(name="next", animated=False, id=965672202416570428)
HOME = discord.PartialEmoji(name="home", animated=False, id=965672202420764713)
PREVIOUS = discord.PartialEmoji(name="previous", animated=False, id=965672202424950795)
MINIMIZE = discord.PartialEmoji(name="minimize", animated=False, id=965672202424963142)
PLAY = discord.PartialEmoji(name="play", animated=False, id=965672202441723994)
EQUALIZER = discord.PartialEmoji(name="equalizer", animated=False, id=965672202454323250)
RANDOM = discord.PartialEmoji(name="random", animated=False, id=965672202458509463)
MENU = discord.PartialEmoji(name="menu", animated=False, id=965672202466910238)
PAUSE = discord.PartialEmoji(name="pause", animated=False, id=965672202466910268)
FASTFORWARD = discord.PartialEmoji(name="fastforward", animated=False, id=965672202513027122)
# noinspection SpellCheckingInspection
VOLUME_UP = discord.PartialEmoji(name="volumeup", animated=False, id=965672202517225492)
STOP = discord.PartialEmoji(name="stop", animated=False, id=965672202563362926)
MUTE = discord.PartialEmoji(name="mute", animated=False, id=965672202592747550)
ARTIST = discord.PartialEmoji(name="artist", animated=False, id=965672202592747600)
SEARCH = discord.PartialEmoji(name="search", animated=False, id=965672202596921384)
# noinspection SpellCheckingInspection
STACKED_ADD_FOLDER = discord.PartialEmoji(name="stackedaddfolder", animated=False, id=965672202785661029)
# noinspection SpellCheckingInspection
MUSICAL_NOTE = discord.PartialEmoji(name="musicalnote", animated=False, id=965674278144077824)
CHECK = discord.PartialEmoji(name="check", animated=False, id=967466875535626260)
TRASH = discord.PartialEmoji(name="trash", animated=False, id=967752655017484318)
URL = discord.PartialEmoji(name="url", animated=False, id=967753966093991968)
CLEAR = discord.PartialEmoji(name="clear", animated=False, id=967756040521252924)
DOWNLOAD = discord.PartialEmoji(name="download", animated=False, id=967760183977730079)
NAME = discord.PartialEmoji(name="name", animated=False, id=967768470781579284)
UPDATE = discord.PartialEmoji(name="update", animated=False, id=967810735851860059)
INFO = discord.PartialEmoji(name="info", animated=False, id=967827814160158820)
QUEUE = discord.PartialEmoji(name="queue", animated=False, id=967902316185415681)
DUPLICATE = discord.PartialEmoji(name="duplicate", animated=False, id=967922711026343966)
SSL = discord.PartialEmoji(name="ssl", animated=False, id=969662672570294342)
HOST = discord.PartialEmoji(name="host", animated=False, id=969675288642486363)
TIMEOUT = discord.PartialEmoji(name="timeout", animated=False, id=969676222697205810)
PORT = discord.PartialEmoji(name="port", animated=False, id=969676962824069200)
PASSWORD = discord.PartialEmoji(name="password", animated=False, id=969690513911459910)
# noinspection SpellCheckingInspection
CLOUD_SERVER = discord.PartialEmoji(name="cloudserver", animated=False, id=972827895900811337)



from __future__ import annotations

import dataclasses
from typing import Annotated

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Timescale:
    speed: Annotated[float | None, ValueRange(min=0.0, max=float("inf"))] = None
    pitch: Annotated[float | None, ValueRange(min=0.0, max=float("inf"))] = None
    rate: Annotated[float | None, ValueRange(min=0.0, max=float("inf"))] = None

    def to_dict(self) -> dict[str, float]:
        response = {}
        if self.speed is not None:
            response["speed"] = self.speed
        if self.pitch is not None:
            response["pitch"] = self.pitch
        if self.rate is not None:
            response["rate"] = self.rate
        return response



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Any

from pylav.constants.node import TRACK_VERSION
from pylav.logging import getLogger
from pylav.utils.vendor.lavalink_py.datarw import DataWriter

if TYPE_CHECKING:
    from pylav.nodes.node import Node

LOGGER = getLogger("PyLav.Track.Decoder")


# noinspection SpellCheckingInspection,PyPep8Naming
def encode_track(
    title: str,
    author: str,
    length: int,
    identifier: str,
    isStream: bool,
    uri: str | None,
    sourceName: str,
    artworkUrl: str | None = None,
    isrc: str | None = None,
    probe: str | None = None,
    version: int = TRACK_VERSION,
    **kwargs: Any,
) -> str:
    writer = DataWriter()
    writer.write_version(version)
    writer.write_utf(title)
    writer.write_utf(author)
    writer.write_long(length)
    writer.write_utf(identifier)
    writer.write_boolean(isStream)
    if version >= 2:
        writer.write_nullable_utf(uri)
    if version >= 3:
        writer.write_nullable_utf(artworkUrl)
        writer.write_nullable_utf(isrc)
    writer.write_utf(sourceName)
    match sourceName:
        case "local" | "http" if probe is not None:
            writer.write_utf(probe)
        case "spotify" | "applemusic" | "deezer" if version == 2:
            writer.write_nullable_utf(isrc)
            writer.write_nullable_utf(artworkUrl)
        case "yandexmusic" if version == 2:
            writer.write_nullable_utf(artworkUrl)
    writer.write_long(0)
    return writer.to_base64()


# noinspection SpellCheckingInspection,PyPep8Naming
async def async_encoder(
    title: str,
    author: str,
    length: int,
    identifier: str,
    is_stream: bool,
    uri: str | None,
    source: str,
    version: int = TRACK_VERSION,
    artworkUrl: str | None = None,
    isrc: str | None = None,
    probe: str | None = None,
    **kwargs: Any,
) -> str:
    return await asyncio.to_thread(
        encode_track,
        title=title,
        author=author,
        length=length,
        identifier=identifier,
        is_stream=is_stream,
        uri=uri,
        source=source,
        artworkUrl=artworkUrl,
        isrc=isrc,
        probe=probe,
        version=version,
        **kwargs,
    )


async def async_re_encoder(track: str, node: Node) -> str:
    track_obj = await node.fetch_decodetrack(track, raise_on_failure=True)
    track_data = track_obj.info.to_dict()
    return await async_encoder(**track_data)


async def async_bulk_re_encoder(track: list[str], node: Node) -> AsyncIterator[str]:
    track_objs = await node.post_decodetracks(track, raise_on_failure=True)
    for track_obj in track_objs:
        track_data = track_obj.info.to_dict()
        yield await async_encoder(**track_data)



from __future__ import annotations

import dataclasses
from typing import NotRequired  # noqa

from piccolo.columns import Column

from pylav.constants.node import TRACK_VERSION
from pylav.nodes.api.responses.shared import TrackPluginInfo
from pylav.storage.database.tables.tracks import TrackRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Info:
    identifier: str
    isSeekable: bool
    author: str
    length: int
    isStream: bool
    position: int
    title: str
    uri: str | None
    sourceName: str | None
    artworkUrl: str | None = None
    isrc: str | None = None
    version: NotRequired[int] = TRACK_VERSION

    def to_dict(self) -> JSON_DICT_TYPE:
        return dataclasses.asdict(self)

    def to_database(self) -> dict[Column, str | None]:
        # noinspection SpellCheckingInspection
        return {
            TrackRow.identifier: self.identifier,
            TrackRow.title: self.title,
            TrackRow.uri: self.uri,
            TrackRow.sourceName: self.sourceName,
            TrackRow.isrc: self.isrc,
            TrackRow.artworkUrl: self.artworkUrl,
        }


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Track:
    info: Info
    encoded: str = None
    pluginInfo: TrackPluginInfo | None = None

    def __post_init__(self) -> None:
        if self.pluginInfo is None:
            object.__setattr__(self, "pluginInfo", TrackPluginInfo(kwargs=None))

    def set_version(self, version: int) -> None:
        object.__setattr__(self.info, "version", version)

    def to_dict(self) -> JSON_DICT_TYPE:
        return {
            "info": dataclasses.asdict(self.info),
            "encoded": self.encoded,
            "pluginInfo": dataclasses.asdict(self.pluginInfo) if self.pluginInfo else None,
        }



from __future__ import annotations

from piccolo.columns import JSONB, UUID, BigInt, Boolean, Float, Integer
from piccolo.table import Table

from pylav.storage.database.tables.misc import DATABASE_ENGINE


class PlayerStateRow(Table, db=DATABASE_ENGINE, tablename="player_state"):
    primary_key = UUID(primary_key=True)
    id = BigInt(index=True, null=False)
    bot = BigInt(index=True, null=False)
    channel_id = BigInt(null=True, default=None)
    volume = Integer(null=False, default=100)
    position = Float(null=False, default=0.0)  # TODO: Should i make this a Numerical column?
    auto_play_playlist_id = BigInt(null=True, default=1)
    forced_channel_id = BigInt(null=True, default=0)
    text_channel_id = BigInt(null=True, default=0)
    notify_channel_id = BigInt(null=True, default=0)

    paused = Boolean(null=False, default=False)
    repeat_current = Boolean(null=False, default=False)
    repeat_queue = Boolean(null=False, default=False)
    shuffle = Boolean(null=False, default=True)
    auto_shuffle = Boolean(null=False, default=False)
    auto_play = Boolean(null=False, default=False)
    playing = Boolean(null=False, default=False)
    effect_enabled = Boolean(null=False, default=False)
    self_deaf = Boolean(null=False, default=False)

    current = JSONB(null=True, default={})
    queue = JSONB(null=False, default=[])
    history = JSONB(null=False, default=[])
    effects = JSONB(null=False, default={})
    extras = JSONB(null=False, default={})



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class PlaylistSaveModal(discord.ui.Modal):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        button: discord.ui.Button,
        title: str,
        timeout: float | None = None,
    ):
        self.cog = cog
        self._button = button
        super().__init__(title=title, timeout=timeout)
        self.text = discord.ui.TextInput(
            style=discord.TextStyle.short,
            label=shorten_string(max_length=100, string=_("Enter the name for the new playlist.")),
            placeholder=shorten_string(max_length=100, string=_("My awesome new playlist")),
            min_length=3,
            max_length=64,
        )
        self.add_item(self.text)

    async def on_submit(self, interaction: DISCORD_INTERACTION_TYPE):
        await self.cog.slash_playlist_save.callback(self.cog, interaction, name=self.text.value.strip())



from __future__ import annotations

import asyncio
from pathlib import Path

from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.extension.red.ui.menus.playlist import PlaylistPickerMenu
from pylav.extension.red.ui.selectors.playlist import PlaylistSelectSelector
from pylav.extension.red.ui.sources.playlist import PlaylistPickerSource
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.playlist import Playlist
from pylav.type_hints.bot import DISCORD_COG_TYPE

_ = Translator("PyLav", Path(__file__))


async def maybe_prompt_for_playlist(
    cog: DISCORD_COG_TYPE, context: PyLavContext, playlists: list[Playlist]
) -> Playlist | None:
    if len(playlists) > 1:
        playlist_picker = PlaylistPickerMenu(
            cog=cog,
            bot=cog.bot,
            source=PlaylistPickerSource(
                guild_id=context.guild.id,
                cog=cog,
                pages=playlists,
                message_str=shorten_string(
                    max_length=100, string=_("Multiple playlists matched. Pick the one which you meant.")
                ),
            ),
            selector_cls=PlaylistSelectSelector,
            delete_after_timeout=True,
            clear_buttons_after=True,
            starting_page=0,
            selector_text=shorten_string(max_length=100, string=_("Pick a playlist.")),
            original_author=context.interaction.user if context.interaction else context.author,
        )

        await playlist_picker.start(context)
        try:
            await playlist_picker.wait_for_response()
            playlist = playlist_picker.result
        except asyncio.TimeoutError:
            playlist = None
    else:
        playlist = playlists[0]

    return playlist



from __future__ import annotations

from pathlib import Path

import discord
from discord import app_commands
from redbot.core import commands
from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.extension.red import errors
from pylav.extension.red.errors import NotDJError, UnauthorizedChannelError
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


def always_hidden(slash: bool = False):
    async def pred(__: DISCORD_INTERACTION_TYPE | PyLavContext) -> bool:
        return False

    return app_commands.check(pred) if slash else commands.check(pred)


def requires_player(slash: bool = False):
    async def pred(context: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(context, discord.Interaction):
            if not context.response.is_done():
                await context.response.defer(ephemeral=True)
            bot = context.client
            _lavalink = getattr(bot, "pylav", None)
            player = _lavalink.get_player(context.guild) if _lavalink else None
        else:
            bot = context.bot
            _lavalink = getattr(bot, "pylav", None)
            player = _lavalink.get_player(context.guild) if _lavalink else None
        if not _lavalink:
            return False
        if not player:
            raise errors.MediaPlayerNotFoundError(context)
        return True

    return app_commands.check(pred) if slash else commands.check(pred)


def can_run_command_in_channel(slash: bool = False):
    async def pred(context: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(context, discord.Interaction):
            if not context.response.is_done():
                await context.response.defer(ephemeral=True)
            bot = context.client
            _lavalink = getattr(bot, "pylav", None)
            player = _lavalink.get_player(context.guild) if _lavalink else None
        else:
            bot = context.bot
            _lavalink = getattr(bot, "pylav", None)
            player = _lavalink.get_player(context.guild) if _lavalink else None
        if not _lavalink:
            return False
        if not context.guild:
            return True
        if player:
            config = player.config
        else:
            config = bot.pylav.player_config_manager.get_config(context.guild.id)

        if (channel_id := await config.fetch_text_channel_id()) != 0 and channel_id != context.channel.id:
            raise UnauthorizedChannelError(channel=channel_id)
        return True

    return app_commands.check(pred) if slash else commands.check(pred)


async def is_dj_logic(
    context: PyLavContext | DISCORD_INTERACTION_TYPE | discord.Message, bot: DISCORD_BOT_TYPE | None = None
) -> bool | None:
    guild = context.guild
    if isinstance(context, discord.Interaction):
        if not context.response.is_done():
            await context.response.defer(ephemeral=True)
        bot = bot or context.client
        author = context.user
    else:
        bot = bot or context.bot
        author = context.author
    return await bot.pylav.is_dj(user=author, guild=guild, additional_role_ids=None, additional_user_ids={*bot.owner_ids, guild.owner_id}, bot=bot) if (getattr(bot, "pylav", None) and guild) else False  # type: ignore


def invoker_is_dj(slash: bool = False):
    async def pred(context: PyLavContext | DISCORD_INTERACTION_TYPE):
        is_dj = await is_dj_logic(context)
        if is_dj is False:
            raise NotDJError(
                context,
            )
        return True

    return app_commands.check(pred) if slash else commands.check(pred)



from __future__ import annotations

from asyncpg import Connection

from pylav.storage.migrations.logging import LOGGER


async def low_level_v_1_3_8_migration(con: Connection) -> None:
    await low_level_v_1_3_8_tracks(con)


async def low_level_v_1_3_8_tracks(con: Connection) -> None:
    await run_tracks_migration_v_1_3_8(con)


async def run_tracks_migration_v_1_3_8(con: Connection) -> None:
    """
    Add the Artwork column to the tracks table.
    """
    version = await con.fetchval("SELECT version from version;")
    if version is None:
        return

    has_column = """
            SELECT EXISTS (SELECT 1
            FROM information_schema.columns
            WHERE table_name='track' AND column_name='artworkUrl')
            """
    has_artwork = await con.fetchval(has_column)
    if not has_artwork:
        LOGGER.info("----------- Migrating Tracks to PyLav 1.3.8 ---------")
        alter_table = """
        ALTER TABLE IF EXISTS track
        ADD COLUMN IF NOT EXISTS "artworkUrl" text COLLATE pg_catalog."default"
        """
        await con.execute(alter_table)



from __future__ import annotations

from dataclasses import dataclass

from pylav.compat import json
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.player_state import PlayerStateRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True)
class PlayerState(CachedModel):
    id: int
    bot: int
    channel_id: int
    volume: int
    position: float
    auto_play_playlist_id: int | None
    text_channel_id: int | None
    notify_channel_id: int | None
    forced_channel_id: int | None

    paused: bool
    repeat_current: bool
    repeat_queue: bool
    shuffle: bool
    auto_shuffle: bool
    auto_play: bool
    playing: bool
    effect_enabled: bool
    self_deaf: bool

    current: JSON_DICT_TYPE | None
    queue: list[JSON_DICT_TYPE]
    history: list[JSON_DICT_TYPE]
    effects: JSON_DICT_TYPE
    extras: JSON_DICT_TYPE
    pk: None = None

    def get_cache_key(self) -> str:
        return f"{self.id}:{self.bot}:{self.channel_id}"

    def __post_init__(self) -> None:
        if isinstance(self.current, str):
            self.current = json.loads(self.current)
        if isinstance(self.queue, str):
            self.queue = json.loads(self.queue)
        if isinstance(self.history, str):
            self.history = json.loads(self.history)
        if isinstance(self.effects, str):
            self.effects = json.loads(self.effects)
        if isinstance(self.extras, str):
            self.extras = json.loads(self.extras)

    async def delete(self) -> None:
        """Delete the player state from the database"""
        await PlayerStateRow.delete().where((PlayerStateRow.id == self.id) & (PlayerStateRow.bot == self.bot))

    async def save(self) -> None:
        """Save the player state to the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerStateRow.raw(
            """
            INSERT INTO player_state (
                id,
                bot,
                channel_id,
                volume,
                position,
                auto_play_playlist_id,
                forced_channel_id,
                text_channel_id,
                notify_channel_id,
                paused,
                repeat_current,
                repeat_queue,
                shuffle,
                auto_shuffle,
                auto_play,
                playing,
                effect_enabled,
                self_deaf,
                current,
                queue,
                history,
                effects,
                extras
            )
            VALUES ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE
                SET channel_id = excluded.channel_id,
                    volume = excluded.volume,
                    position = excluded.position,
                    auto_play_playlist_id = excluded.auto_play_playlist_id,
                    forced_channel_id = excluded.forced_channel_id,
                    text_channel_id = excluded.text_channel_id,
                    notify_channel_id = excluded.notify_channel_id,
                    paused = excluded.paused,
                    repeat_current = excluded.repeat_current,
                    repeat_queue = excluded.repeat_queue,
                    shuffle = excluded.shuffle,
                    auto_shuffle = excluded.auto_shuffle,
                    auto_play = excluded.auto_play,
                    playing = excluded.playing,
                    effect_enabled = excluded.effect_enabled,
                    self_deaf = excluded.self_deaf,
                    current = excluded.current,
                    queue = excluded.queue,
                    history = excluded.history,
                    effects = excluded.effects,
                    extras = excluded.extras;
            """,
            self.id,
            self.bot,
            self.channel_id,
            self.volume,
            self.position,
            self.auto_play_playlist_id,
            self.forced_channel_id,
            self.text_channel_id,
            self.notify_channel_id,
            self.paused,
            self.repeat_current,
            self.repeat_queue,
            self.shuffle,
            self.auto_shuffle,
            self.auto_play,
            self.playing,
            self.effect_enabled,
            self.self_deaf,
            json.dumps(self.current),
            json.dumps(self.queue),
            json.dumps(self.history),
            json.dumps(self.effects),
            json.dumps(self.extras),
        )

    @classmethod
    async def get(cls, bot_id: int, guild_id: int) -> PlayerState | None:
        """Get the player state from the database.

        Parameters
        ----------
        bot_id : int
            The bot ID.
        guild_id : int
            The guild ID.

        Returns
        -------
        PlayerState | None
            The player state if found, otherwise None.
        """
        player = (
            await PlayerStateRow.select(
                PlayerStateRow.id,
                PlayerStateRow.bot,
                PlayerStateRow.channel_id,
                PlayerStateRow.volume,
                PlayerStateRow.position,
                PlayerStateRow.auto_play_playlist_id,
                PlayerStateRow.forced_channel_id,
                PlayerStateRow.text_channel_id,
                PlayerStateRow.notify_channel_id,
                PlayerStateRow.paused,
                PlayerStateRow.repeat_current,
                PlayerStateRow.repeat_queue,
                PlayerStateRow.shuffle,
                PlayerStateRow.auto_shuffle,
                PlayerStateRow.auto_play,
                PlayerStateRow.playing,
                PlayerStateRow.effect_enabled,
                PlayerStateRow.self_deaf,
                PlayerStateRow.current,
                PlayerStateRow.queue,
                PlayerStateRow.history,
                PlayerStateRow.effects,
                PlayerStateRow.extras,
            )
            .where((PlayerStateRow.id == guild_id) & (PlayerStateRow.bot == bot_id))
            .first()
            .output(load_json=True, nested=True)
        )
        return cls(**player) if player else None



from __future__ import annotations

import os

from piccolo.engine import PostgresEngine, SQLiteEngine

from pylav.constants.config import POSTGRES_DATABASE, POSTGRES_HOST, POSTGRES_PASSWORD, POSTGRES_PORT, POSTGRES_USER
from pylav.logging import getLogger

_CONFIG = {
    "host": POSTGRES_HOST,
    "port": int(POSTGRES_PORT) if isinstance(POSTGRES_PORT, str) else POSTGRES_PORT,
    "database": POSTGRES_DATABASE,
    "user": POSTGRES_USER,
    "password": POSTGRES_PASSWORD,
}
LOGGER = getLogger("PyLav.Postgres")
DATABASE_ENGINE: PostgresEngine | SQLiteEngine
if os.getenv("PYLAV__SQL", False):
    DATABASE_ENGINE = SQLiteEngine()
    IS_POSTGRES = False
else:
    LOGGER.verbose("Connecting to Postgres server using %r", _CONFIG)
    # noinspection SpellCheckingInspection
    DATABASE_ENGINE = PostgresEngine(
        config=_CONFIG,
        extensions=(
            "uuid-ossp",
            "btree_gin",
        ),
    )
    IS_POSTGRES = True



from __future__ import annotations

import decimal
import errno
import os
from typing import Any

import aiopath

from pylav.extension.m3u.mixins import BasePathMixin, GroupedBasePathMixin
from pylav.extension.m3u.parser import format_date_time, parse
from pylav.extension.m3u.protocols import EXT_X_KEY, EXT_X_MAP, EXT_X_SESSION_KEY, EXT_X_START


class MalformedPlaylistError(Exception):
    pass


class M3U8:
    __slots__ = (
        "data",
        "_base_uri",
        "keys",
        "segments",
        "files",
        "media",
        "playlists",
        "iframe_playlists",
        "segment_map",
        "start",
        "server_control",
        "part_inf",
        "skip",
        "rendition_reports",
        "session_data",
        "session_keys",
        "preload_hint",
        "content_steering",
        "_base_path",
        "is_variant",
        "is_independent_segments",
        "media_sequence",
        "discontinuity_sequence",
        "allow_cache",
        "version",
        "target_duration",
        "playlist_type",
        "is_i_frames_only",
        "is_endlist",
    )

    simple_attributes = (
        # obj attribute      # parser attribute
        ("is_variant", "is_variant"),
        ("is_endlist", "is_endlist"),
        ("is_i_frames_only", "is_i_frames_only"),
        ("target_duration", "targetduration"),
        ("media_sequence", "media_sequence"),
        ("program_date_time", "program_date_time"),
        ("is_independent_segments", "is_independent_segments"),
        ("version", "version"),
        ("allow_cache", "allow_cache"),
        ("playlist_type", "playlist_type"),
        ("discontinuity_sequence", "discontinuity_sequence"),
    )

    def __init__(
        self,
        content: str = None,
        base_path: str = None,
        base_uri: str = None,
        strict: bool = False,
        custom_tags_parser=None,
    ) -> None:
        self.is_variant = False

        if content is not None:
            self.data = parse(content, strict, custom_tags_parser)
        else:
            self.data = {}
        self._base_uri = base_uri
        if self._base_uri and not self._base_uri.endswith("/"):
            self._base_uri += "/"

        self._initialize_attributes()
        self.base_path = base_path

    def _initialize_attributes(self) -> None:
        self.keys = [Key(base_uri=self.base_uri, **params) if params else None for params in self.data.get("keys", [])]
        self.segments = SegmentList(
            [
                Segment(base_uri=self.base_uri, keyobject=find_key(segment.get("key", {}), self.keys), **segment)
                for segment in self.data.get("segments", [])
            ]
        )
        # self.keys = get_uniques([ segment.key for segment in self.segments ])
        for attr, param in self.simple_attributes:
            setattr(self, attr, self.data.get(param))

        self.files = [key.uri for key in self.keys if key and key.uri not in self.files]

        self.files.extend(self.segments.uri)

        self.media = MediaList([Media(base_uri=self.base_uri, **media) for media in self.data.get("media", [])])

        self.playlists = PlaylistList(
            [
                Playlist(base_uri=self.base_uri, media=self.media, **playlist)
                for playlist in self.data.get("playlists", [])
            ]
        )

        self.iframe_playlists = PlaylistList()
        for ifr_pl in self.data.get("iframe_playlists", []):
            self.iframe_playlists.append(
                IFramePlaylist(
                    base_uri=self.base_uri, uri=ifr_pl["uri"], iframe_stream_info=ifr_pl["iframe_stream_info"]
                )
            )
        self.segment_map = self.data.get("segment_map")

        start = self.data.get("start", None)
        self.start = start and Start(**start)

        server_control = self.data.get("server_control", None)
        self.server_control = server_control and ServerControl(**server_control)

        part_inf = self.data.get("part_inf", None)
        self.part_inf = part_inf and PartInformation(**part_inf)

        skip = self.data.get("skip", None)
        self.skip = skip and Skip(**skip)

        self.rendition_reports = RenditionReportList(
            [
                RenditionReport(base_uri=self.base_uri, **rendition_report)
                for rendition_report in self.data.get("rendition_reports", [])
            ]
        )

        self.session_data = SessionDataList(
            [
                SessionData(**session_data)
                for session_data in self.data.get("session_data", [])
                if "data_id" in session_data
            ]
        )

        self.session_keys = [
            SessionKey(base_uri=self.base_uri, **params) if params else None
            for params in self.data.get("session_keys", [])
        ]

        preload_hint = self.data.get("preload_hint", None)
        self.preload_hint = preload_hint and PreloadHint(base_uri=self.base_uri, **preload_hint)

        content_steering = self.data.get("content_steering", None)
        self.content_steering = content_steering and ContentSteering(base_uri=self.base_uri, **content_steering)

    def __unicode__(self) -> str:
        return self.dumps()

    @property
    def base_uri(self) -> str:
        return self._base_uri

    @base_uri.setter
    def base_uri(self, new_base_uri: str | None) -> None:
        self._base_uri = new_base_uri
        self.media.base_uri = new_base_uri
        self.playlists.base_uri = new_base_uri
        self.iframe_playlists.base_uri = new_base_uri
        self.segments.base_uri = new_base_uri
        self.rendition_reports.base_uri = new_base_uri
        for key in self.keys:
            if key:
                key.base_uri = new_base_uri
        for key in self.session_keys:
            if key:
                key.base_uri = new_base_uri
        if self.preload_hint:
            self.preload_hint.base_uri = new_base_uri
        if self.content_steering:
            self.content_steering.base_uri = new_base_uri

    @property
    def base_path(self) -> str:
        return self._base_path

    @base_path.setter
    def base_path(self, newbase_path: str):
        self._base_path = newbase_path
        self._update_base_path()

    def _update_base_path(self) -> None:
        if self._base_path is None:
            return
        for key in self.keys:
            if key:
                key.base_path = self._base_path
        for key in self.session_keys:
            if key:
                key.base_path = self._base_path
        self.media.base_path = self._base_path
        self.segments.base_path = self._base_path
        self.playlists.base_path = self._base_path
        self.iframe_playlists.base_path = self._base_path
        self.rendition_reports.base_path = self._base_path
        if self.preload_hint:
            self.preload_hint.base_path = self._base_path
        if self.content_steering:
            self.content_steering.base_path = self._base_path

    def add_playlist(self, playlist: Playlist) -> None:
        self.is_variant = True
        self.playlists.append(playlist)

    def add_iframe_playlist(self, iframe_playlist: IFramePlaylist | None) -> None:
        if iframe_playlist is not None:
            self.is_variant = True
            self.iframe_playlists.append(iframe_playlist)

    def add_media(self, media: Media) -> None:
        self.media.append(media)

    def add_segment(self, segment: Segment) -> None:
        self.segments.append(segment)

    def add_rendition_report(self, report: RenditionReport) -> None:
        self.rendition_reports.append(report)

    def dumps(self) -> str:
        """
        Returns the current m3u8 as a string.
        You could also use unicode(<this obj>) or str(<this obj>)
        """
        output = ["#EXTM3U"]

        if self.content_steering:
            output.append(str(self.content_steering))
        if self.is_independent_segments:
            output.append("#EXT-X-INDEPENDENT-SEGMENTS")
        if self.media_sequence:
            output.append(f"#EXT-X-MEDIA-SEQUENCE:{str(self.media_sequence)}")
        if self.discontinuity_sequence:
            output.append(f"#EXT-X-DISCONTINUITY-SEQUENCE:{number_to_string(self.discontinuity_sequence)}")
        if self.allow_cache:
            output.append(f"#EXT-X-ALLOW-CACHE:{self.allow_cache.upper()}")
        if self.version:
            output.append(f"#EXT-X-VERSION:{str(self.version)}")
        if self.target_duration:
            output.append(f"#EXT-X-TARGETDURATION:{number_to_string(self.target_duration)}")
        if self.playlist_type is not None and self.playlist_type != "":
            output.append(f"#EXT-X-PLAYLIST-TYPE:{str(self.playlist_type).upper()}")

        if self.start:
            output.append(str(self.start))
        if self.is_i_frames_only:
            output.append("#EXT-X-I-FRAMES-ONLY")
        if self.server_control:
            output.append(str(self.server_control))
        if self.is_variant:
            if self.media:
                output.append(str(self.media))
            output.append(str(self.playlists))
            if self.iframe_playlists:
                output.append(str(self.iframe_playlists))
        if self.part_inf:
            output.append(str(self.part_inf))
        if self.skip:
            output.append(str(self.skip))
        if self.session_data:
            output.append(str(self.session_data))

        output.extend(str(key) for key in self.session_keys)
        output.append(str(self.segments))

        if self.preload_hint:
            output.append(str(self.preload_hint))

        if self.rendition_reports:
            output.append(str(self.rendition_reports))

        if self.is_endlist:
            output.append("#EXT-X-ENDLIST")

        # ensure that the last line is terminated correctly
        if output[-1] and not output[-1].endswith("\n"):
            output.append("")

        return "\n".join(output)

    async def dump(self, filename: str) -> None:
        """
        Saves the current m3u8 to ``filename``
        """
        self._create_sub_directories(filename)
        file = aiopath.AsyncPath(filename)
        async with file.open("w") as fileobj:
            await fileobj.write(self.dumps())

    @staticmethod
    def _create_sub_directories(filename: str) -> None:
        basename = os.path.dirname(filename)
        try:
            if basename:
                os.makedirs(basename)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise


class Segment(BasePathMixin):
    """
    A video segment from a M3U8 playlist

    `uri`
      a string with the segment uri

    `title`
      title attribute from EXTINF parameter

    `program_date_time`
      Returns the EXT-X-PROGRAM-DATE-TIME as a datetime. This field is only set
      if EXT-X-PROGRAM-DATE-TIME exists for this segment
      http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.5

    `current_program_date_time`
      Returns a datetime of this segment, either the value of `program_date_time`
      when EXT-X-PROGRAM-DATE-TIME is set or a calculated value based on previous
      segments' EXT-X-PROGRAM-DATE-TIME and EXTINF values

    `discontinuity`
      Returns a boolean indicating if a EXT-X-DISCONTINUITY tag exists
      http://tools.ietf.org/html/draft-pantos-http-live-streaming-13#section-3.4.11

    `cue_out_start`
      Returns a boolean indicating if a EXT-X-CUE-OUT tag exists

    `cue_out`
      Returns a boolean indicating if a EXT-X-CUE-OUT-CONT tag exists
      Note: for backwards compatibility, this will be True when cue_out_start
            is True, even though this tag did not exist in the input, and
            EXT-X-CUE-OUT-CONT will not exist in the output

    `cue_in`
      Returns a boolean indicating if a EXT-X-CUE-IN tag exists

    `scte35`
      Base64 encoded SCTE35 metadata if available

    `scte35_duration`
      Planned SCTE35 duration

    `duration`
      duration attribute from EXTINF parameter

    `base_uri`
      uri the key comes from in URI hierarchy. ex.: http://example.com/path/to

    `bitrate`
      bitrate attribute from EXT-X-BITRATE parameter

    `byterange`
      byterange attribute from EXT-X-BYTERANGE parameter

    `key`
      Key used to encrypt the segment (EXT-X-KEY)

    `parts`
      partial segments that make up this segment

    `dateranges`
      any dateranges that should precede the segment

    `gap_tag`
      GAP tag indicates that a Media Segment is missing

    `custom_parser_values`
        Additional values which custom_tags_parser might store per segment
    """

    def __init__(
        self,
        uri=None,
        base_uri=None,
        program_date_time=None,
        current_program_date_time=None,
        duration=None,
        title=None,
        bitrate=None,
        byterange=None,
        cue_out=False,
        cue_out_start=False,
        cue_in=False,
        discontinuity=False,
        key=None,  # noqa
        scte35=None,
        scte35_duration=None,
        keyobject=None,
        parts=None,
        init_section=None,
        dateranges=None,
        gap_tag=None,
        custom_parser_values=None,
    ):
        self.uri = uri
        self.duration = duration
        self.title = title
        self._base_uri = base_uri
        self.bitrate = bitrate
        self.byterange = byterange
        self.program_date_time = program_date_time
        self.current_program_date_time = current_program_date_time
        self.discontinuity = discontinuity
        self.cue_out_start = cue_out_start
        self.cue_out = cue_out
        self.cue_in = cue_in
        self.scte35 = scte35
        self.scte35_duration = scte35_duration
        self.key = keyobject
        self.parts = PartialSegmentList(
            [PartialSegment(base_uri=self._base_uri, **partial) for partial in parts] if parts else []
        )
        if init_section is not None:
            self.init_section = InitializationSection(self._base_uri, **init_section)
        else:
            self.init_section = None
        self.dateranges = DateRangeList([DateRange(**daterange) for daterange in dateranges] if dateranges else [])
        self.gap_tag = gap_tag
        self.custom_parser_values = custom_parser_values or {}

        # Key(base_uri=base_uri, **key) if key else None

    def add_part(self, part: PartInformation) -> None:
        self.parts.append(part)

    def dumps(self, last_segment: Segment | None) -> str:
        output = []

        if (
            (not last_segment or self.key == last_segment.key)
            and self.key
            and last_segment is None
            or last_segment
            and self.key != last_segment.key
        ):
            output.extend((str(self.key), "\n"))
        if last_segment and self.init_section != last_segment.init_section and not self.init_section:
            raise MalformedPlaylistError("init section can't be None if previous is not None")
        elif (
            last_segment
            and self.init_section != last_segment.init_section
            or self.init_section
            and last_segment is None
        ):
            output.extend((str(self.init_section), "\n"))
        if self.discontinuity:
            output.append("#EXT-X-DISCONTINUITY\n")
        if self.program_date_time:
            output.append(f"#EXT-X-PROGRAM-DATE-TIME:{format_date_time(self.program_date_time)}\n")

        if len(self.dateranges):
            output.extend((str(self.dateranges), "\n"))
        if self.cue_out_start:
            output.append(f"#EXT-X-CUE-OUT{f':{self.scte35_duration}' if self.scte35_duration else ''}\n")

        elif self.cue_out:
            output.append("#EXT-X-CUE-OUT-CONT\n")
        if self.cue_in:
            output.append("#EXT-X-CUE-IN\n")

        if self.parts:
            output.extend((str(self.parts), "\n"))
        if self.uri:
            if self.duration is not None:
                output.append(f"#EXTINF:{number_to_string(self.duration)},")
                if self.title:
                    output.append(self.title)
                output.append("\n")

            if self.byterange:
                output.append(f"#EXT-X-BYTERANGE:{self.byterange}\n")

            if self.bitrate:
                output.append(f"#EXT-X-BITRATE:{self.bitrate}\n")

            if self.gap_tag:
                output.append("#EXT-X-GAP\n")

            output.append(self.uri)

        return "".join(output)

    def __str__(self) -> str:
        return self.dumps(None)

    @property
    def base_path(self) -> str | None:
        return super().base_path

    @base_path.setter
    def base_path(self, newbase_path: str | None) -> None:
        super(Segment, self.__class__).base_path.fset(self, newbase_path)  # type: ignore
        self.parts.base_path = newbase_path
        if self.init_section is not None:
            self.init_section.base_path = newbase_path

    @property
    def base_uri(self) -> str | None:
        return self._base_uri

    @base_uri.setter
    def base_uri(self, newbase_uri: str | None) -> None:
        self._base_uri = newbase_uri
        self.parts.base_uri = newbase_uri
        if self.init_section is not None:
            self.init_section.base_uri = newbase_uri


class SegmentList(list, GroupedBasePathMixin):
    def __str__(self) -> str:
        output = []
        last_segment = None
        for segment in self:
            output.append(segment.dumps(last_segment))
            last_segment = segment
        return "\n".join(output)

    @property
    def uri(self) -> list[str]:
        return [seg.uri for seg in self]

    def by_key(self, key: Key) -> list[Segment]:
        return [segment for segment in self if segment.key == key]


class PartialSegment(BasePathMixin):
    __slots__ = (
        "uri",
        "base_uri",
        "duration",
        "program_date_time",
        "current_program_date_time",
        "byterange",
        "independent",
        "gap",
        "dateranges",
        "gap_tag",
    )

    def __init__(
        self,
        base_uri,
        uri,
        duration,
        program_date_time=None,
        current_program_date_time=None,
        byterange=None,
        independent=None,
        gap=None,
        dateranges=None,
        gap_tag=None,
    ) -> None:
        self.base_uri = base_uri
        self.uri = uri
        self.duration = duration
        self.program_date_time = program_date_time
        self.current_program_date_time = current_program_date_time
        self.byterange = byterange
        self.independent = independent
        self.gap = gap
        self.dateranges = DateRangeList([DateRange(**daterange) for daterange in dateranges] if dateranges else [])
        self.gap_tag = gap_tag

    def dumps(self, last_segment: Segment | None) -> str:  # noqa
        output = []

        if len(self.dateranges):
            output.extend((str(self.dateranges), "\n"))
        if self.gap_tag:
            output.append("#EXT-X-GAP\n")

        output.append(f'#EXT-X-PART:DURATION={number_to_string(self.duration)},URI="{self.uri}"')

        if self.independent:
            output.append(f",INDEPENDENT={self.independent}")

        if self.byterange:
            output.append(f",BYTERANGE={self.byterange}")

        if self.gap:
            output.append(f",GAP={self.gap}")

        return "".join(output)

    def __str__(self) -> str:
        return self.dumps(None)


class PartialSegmentList(list, GroupedBasePathMixin):
    def __str__(self) -> str:
        output = [str(part) for part in self]
        return "\n".join(output)


class Key(BasePathMixin):
    tag = EXT_X_KEY

    def __init__(self, method, base_uri, uri=None, iv=None, keyformat=None, keyformatversions=None, **kwargs):
        self.method = method
        self.uri = uri
        self.iv = iv
        self.keyformat = keyformat
        self.keyformatversions = keyformatversions
        self.base_uri = base_uri
        self._extra_params = kwargs

    def __str__(self) -> str:
        output = [
            f"METHOD={self.method}",
        ]
        if self.uri:
            output.append(f'URI="{self.uri}"')
        if self.iv:
            output.append(f"IV={self.iv}")
        if self.keyformat:
            output.append(f'KEYFORMAT="{self.keyformat}"')
        if self.keyformatversions:
            output.append(f'KEYFORMATVERSIONS="{self.keyformatversions}"')

        return f"{self.tag}:{','.join(output)}"

    def __eq__(self, other: Key) -> bool:
        return (
            (
                self.method == other.method
                and self.uri == other.uri
                and self.iv == other.iv
                and self.base_uri == other.base_uri
                and self.keyformat == other.keyformat
                and self.keyformatversions == other.keyformatversions
            )
            if other
            else False
        )

    def __ne__(self, other: Key) -> bool:
        return not self.__eq__(other)


class InitializationSection(BasePathMixin):
    tag = EXT_X_MAP

    def __init__(self, base_uri, uri, byterange=None) -> None:
        self.base_uri = base_uri
        self.uri = uri
        self.byterange = byterange

    def __str__(self) -> str:
        output = []
        if self.uri:
            output.append(f"URI={quoted(self.uri)}")
        if self.byterange:
            output.append(f"BYTERANGE={self.byterange}")
        return f"{self.tag}:{','.join(output)}"

    def __eq__(self, other: InitializationSection) -> bool:
        return (
            self.uri == other.uri and self.byterange == other.byterange and self.base_uri == other.base_uri
            if other
            else False
        )

    def __ne__(self, other: InitializationSection) -> bool:
        return not self.__eq__(other)


class SessionKey(Key):
    tag = EXT_X_SESSION_KEY


class Playlist(BasePathMixin):
    def __init__(self, uri, stream_info, media, base_uri) -> None:
        self.uri = uri
        self.base_uri = base_uri

        if (resolution := stream_info.get("resolution")) is not None:
            resolution = resolution.strip('"')
            values = resolution.split("x")
            resolution_pair = (int(values[0]), int(values[1]))
        else:
            resolution_pair = None

        self.stream_info = StreamInfo(
            bandwidth=stream_info["bandwidth"],
            video=stream_info.get("video"),
            audio=stream_info.get("audio"),
            subtitles=stream_info.get("subtitles"),
            closed_captions=stream_info.get("closed_captions"),
            average_bandwidth=stream_info.get("average_bandwidth"),
            program_id=stream_info.get("program_id"),
            resolution=resolution_pair,
            codecs=stream_info.get("codecs"),
            frame_rate=stream_info.get("frame_rate"),
            video_range=stream_info.get("video_range"),
            hdcp_level=stream_info.get("hdcp_level"),
            pathway_id=stream_info.get("pathway_id"),
        )
        self.media = []
        for media_type in ("audio", "video", "subtitles"):
            if group_id := stream_info.get(media_type):
                self.media += filter(lambda m: m.group_id == group_id, media)

    def __str__(self) -> str:
        media_types = []
        stream_inf = [str(self.stream_info)]
        for media in self.media:
            if media.type in media_types:
                continue
            media_types += [media.type]
            media_type = media.type.upper()
            stream_inf.append(f'{media_type}="{media.group_id}"')

        return f"#EXT-X-STREAM-INF:{','.join(stream_inf)}\n{self.uri}"


class IFramePlaylist(BasePathMixin):
    def __init__(self, base_uri, uri, iframe_stream_info) -> None:
        self.uri = uri
        self.base_uri = base_uri

        if (resolution := iframe_stream_info.get("resolution")) is not None:
            values = resolution.split("x")
            resolution_pair = (int(values[0]), int(values[1]))
        else:
            resolution_pair = None

        self.iframe_stream_info = StreamInfo(
            bandwidth=iframe_stream_info.get("bandwidth"),
            average_bandwidth=iframe_stream_info.get("average_bandwidth"),
            video=iframe_stream_info.get("video"),
            # Audio, subtitles, and closed captions should not exist in
            # EXT-X-I-FRAME-STREAM-INF, so just hardcode them to None.
            audio=None,
            subtitles=None,
            closed_captions=None,
            program_id=iframe_stream_info.get("program_id"),
            resolution=resolution_pair,
            codecs=iframe_stream_info.get("codecs"),
            video_range=iframe_stream_info.get("video_range"),
            hdcp_level=iframe_stream_info.get("hdcp_level"),
            frame_rate=None,
            pathway_id=iframe_stream_info.get("pathway_id"),
        )

    def __str__(self) -> str:
        iframe_stream_inf = []
        if self.iframe_stream_info.program_id:
            iframe_stream_inf.append(f"PROGRAM-ID={self.iframe_stream_info.program_id:d}")
        if self.iframe_stream_info.bandwidth:
            iframe_stream_inf.append(f"BANDWIDTH={self.iframe_stream_info.bandwidth:d}")
        if self.iframe_stream_info.average_bandwidth:
            iframe_stream_inf.append(f"AVERAGE-BANDWIDTH={self.iframe_stream_info.average_bandwidth:d}")
        if self.iframe_stream_info.resolution:
            res = f"{self.iframe_stream_info.resolution[0]}x{self.iframe_stream_info.resolution[1]}"
            iframe_stream_inf.append(f"RESOLUTION={res}")
        if self.iframe_stream_info.codecs:
            iframe_stream_inf.append(f"CODECS={quoted(self.iframe_stream_info.codecs)}")
        if self.iframe_stream_info.video_range:
            iframe_stream_inf.append(f"VIDEO-RANGE={self.iframe_stream_info.video_range}")
        if self.iframe_stream_info.hdcp_level:
            iframe_stream_inf.append(f"HDCP-LEVEL={self.iframe_stream_info.hdcp_level}")
        if self.uri:
            iframe_stream_inf.append(f"URI={quoted(self.uri)}")
        if self.iframe_stream_info.pathway_id:
            iframe_stream_inf.append(f"PATHWAY-ID={quoted(self.iframe_stream_info.pathway_id)}")

        return f"#EXT-X-I-FRAME-STREAM-INF:{','.join(iframe_stream_inf)}"


class StreamInfo:
    bandwidth = None
    closed_captions = None
    average_bandwidth = None
    program_id = None
    resolution = None
    codecs = None
    audio = None
    video = None
    subtitles = None
    frame_rate = None
    video_range = None
    hdcp_level = None
    pathway_id = None

    def __init__(self, **kwargs: Any) -> None:
        self.bandwidth = kwargs.get("bandwidth")
        self.closed_captions = kwargs.get("closed_captions")
        self.average_bandwidth = kwargs.get("average_bandwidth")
        self.program_id = kwargs.get("program_id")
        self.resolution = kwargs.get("resolution")
        self.codecs = kwargs.get("codecs")
        self.audio = kwargs.get("audio")
        self.video = kwargs.get("video")
        self.subtitles = kwargs.get("subtitles")
        self.frame_rate = kwargs.get("frame_rate")
        self.video_range = kwargs.get("video_range")
        self.hdcp_level = kwargs.get("hdcp_level")
        self.pathway_id = kwargs.get("pathway_id")

    def __str__(self) -> str:
        stream_inf = []
        if self.program_id is not None:
            stream_inf.append(f"PROGRAM-ID={self.program_id:d}")
        if self.closed_captions is not None:
            stream_inf.append(f"CLOSED-CAPTIONS={self.closed_captions}")
        if self.bandwidth is not None:
            stream_inf.append(f"BANDWIDTH={self.bandwidth:d}")
        if self.average_bandwidth is not None:
            stream_inf.append(f"AVERAGE-BANDWIDTH={self.average_bandwidth:d}")
        if self.resolution is not None:
            res = f"{self.resolution[0]}x{self.resolution[1]}"
            stream_inf.append(f"RESOLUTION={res}")
        if self.frame_rate is not None:
            stream_inf.append(f"FRAME-RATE={decimal.Decimal(self.frame_rate).quantize(decimal.Decimal('1.000')):g}")
        if self.codecs is not None:
            stream_inf.append(f"CODECS={quoted(self.codecs)}")
        if self.video_range is not None:
            stream_inf.append(f"VIDEO-RANGE={self.video_range}")
        if self.hdcp_level is not None:
            stream_inf.append(f"HDCP-LEVEL={self.hdcp_level}")
        if self.pathway_id is not None:
            stream_inf.append(f"PATHWAY-ID={quoted(self.pathway_id)}")
        return ",".join(stream_inf)


class Media(BasePathMixin):
    def __init__(
        self,
        uri=None,
        type=None,  # noqa
        group_id=None,
        language=None,
        name=None,
        default=None,
        autoselect=None,
        forced=None,
        characteristics=None,
        channels=None,
        assoc_language=None,
        instream_id=None,
        base_uri=None,
        **extras,
    ) -> None:
        self.base_uri = base_uri
        self.uri = uri
        self.type = type
        self.group_id = group_id
        self.language = language
        self.name = name
        self.default = default
        self.autoselect = autoselect
        self.forced = forced
        self.assoc_language = assoc_language
        self.instream_id = instream_id
        self.characteristics = characteristics
        self.channels = channels
        self.extras = extras

    def dumps(self) -> str:
        media_out = []

        if self.uri:
            media_out.append(f"URI={quoted(self.uri)}")
        if self.type:
            media_out.append(f"TYPE={self.type}")
        if self.group_id:
            media_out.append(f"GROUP-ID={quoted(self.group_id)}")
        if self.language:
            media_out.append(f"LANGUAGE={quoted(self.language)}")
        if self.assoc_language:
            media_out.append(f"ASSOC-LANGUAGE={quoted(self.assoc_language)}")
        if self.name:
            media_out.append(f"NAME={quoted(self.name)}")
        if self.default:
            media_out.append(f"DEFAULT={self.default}")
        if self.autoselect:
            media_out.append(f"AUTOSELECT={self.autoselect}")
        if self.forced:
            media_out.append(f"FORCED={self.forced}")
        if self.instream_id:
            media_out.append(f"INSTREAM-ID={quoted(self.instream_id)}")
        if self.characteristics:
            media_out.append(f"CHARACTERISTICS={quoted(self.characteristics)}")
        if self.channels:
            media_out.append(f"CHANNELS={quoted(self.channels)}")

        return f'#EXT-X-MEDIA:{",".join(media_out)}'

    def __str__(self) -> str:
        return self.dumps()


class TagList(list):
    def __str__(self) -> str:
        output = [str(tag) for tag in self]
        return "\n".join(output)


class MediaList(TagList, GroupedBasePathMixin):
    @property
    def uri(self) -> list[str]:
        return [media.uri for media in self]


class PlaylistList(TagList, GroupedBasePathMixin):
    pass


class SessionDataList(TagList):
    pass


class Start:
    def __init__(self, time_offset, precise=None) -> None:
        self.time_offset = float(time_offset)
        self.precise = precise

    def __str__(self) -> str:
        output = [f"TIME-OFFSET={str(self.time_offset)}"]
        if self.precise and self.precise in ["YES", "NO"]:
            output.append(f"PRECISE={str(self.precise)}")

        return f"{EXT_X_START}:{','.join(output)}"


class RenditionReport(BasePathMixin):
    def __init__(self, base_uri, uri, last_msn, last_part=None) -> None:
        self.base_uri = base_uri
        self.uri = uri
        self.last_msn = last_msn
        self.last_part = last_part

    def dumps(self) -> str:
        report = [f"URI={quoted(self.uri)}", f"LAST-MSN={number_to_string(self.last_msn)}"]
        if self.last_part is not None:
            report.append(f"LAST-PART={number_to_string(self.last_part)}")

        return f'#EXT-X-RENDITION-REPORT:{",".join(report)}'

    def __str__(self) -> str:
        return self.dumps()


class RenditionReportList(list, GroupedBasePathMixin):
    def __str__(self) -> str:
        output = [str(report) for report in self]
        return "\n".join(output)


class ServerControl:
    def __init__(
        self, can_skip_until=None, can_block_reload=None, hold_back=None, part_hold_back=None, can_skip_dateranges=None
    ) -> None:
        self.can_skip_until = can_skip_until
        self.can_block_reload = can_block_reload
        self.hold_back = hold_back
        self.part_hold_back = part_hold_back
        self.can_skip_dateranges = can_skip_dateranges

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)

    def dumps(self) -> str:
        ctrl = []
        if self.can_block_reload:
            ctrl.append(f"CAN-BLOCK-RELOAD={self.can_block_reload}")

        ctrl.extend(
            f"{denormalize_attribute(attr)}={number_to_string(self[attr])}"
            for attr in ["hold_back", "part_hold_back"]
            if self[attr]
        )

        if self.can_skip_until:
            ctrl.append(f"CAN-SKIP-UNTIL={number_to_string(self.can_skip_until)}")
            if self.can_skip_dateranges:
                ctrl.append(f"CAN-SKIP-DATERANGES={self.can_skip_dateranges}")

        return f'#EXT-X-SERVER-CONTROL:{",".join(ctrl)}'

    def __str__(self) -> str:
        return self.dumps()


class Skip:
    def __init__(self, skipped_segments: int, recently_removed_dateranges=None) -> None:
        self.skipped_segments = skipped_segments
        self.recently_removed_dateranges = recently_removed_dateranges

    def dumps(self) -> str:
        skip = [f"SKIPPED-SEGMENTS={number_to_string(self.skipped_segments)}"]
        if self.recently_removed_dateranges is not None:
            skip.append(f"RECENTLY-REMOVED-DATERANGES={quoted(self.recently_removed_dateranges)}")

        return f'#EXT-X-SKIP:{",".join(skip)}'

    def __str__(self) -> str:
        return self.dumps()


class PartInformation:
    __slots__ = ("part_target",)

    def __init__(self, part_target=None) -> None:
        self.part_target = part_target

    def dumps(self) -> str:
        return f"#EXT-X-PART-INF:PART-TARGET={number_to_string(self.part_target)}"

    def __str__(self) -> str:
        return self.dumps()


class PreloadHint(BasePathMixin):
    def __init__(self, type, base_uri, uri, byterange_start=None, byterange_length=None) -> None:  # noqa
        self.hint_type = type
        self.base_uri = base_uri
        self.uri = uri
        self.byterange_start = byterange_start
        self.byterange_length = byterange_length

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)

    def dumps(self) -> str:
        hint = [f"TYPE={self.hint_type}", f"URI={quoted(self.uri)}"]

        for attr in ["byterange_start", "byterange_length"]:
            if self[attr] is not None:
                hint.append(f"{denormalize_attribute(attr)}={number_to_string(self[attr])}")

        return f'#EXT-X-PRELOAD-HINT:{",".join(hint)}'

    def __str__(self) -> str:
        return self.dumps()


class SessionData:
    def __init__(self, data_id, value=None, uri=None, language=None) -> None:
        self.data_id = data_id
        self.value = value
        self.uri = uri
        self.language = language

    def dumps(self) -> str:
        session_data_out = [f"DATA-ID={quoted(self.data_id)}"]

        if self.value:
            session_data_out.append(f"VALUE={quoted(self.value)}")
        elif self.uri:
            session_data_out.append(f"URI={quoted(self.uri)}")
        if self.language:
            session_data_out.append(f"LANGUAGE={quoted(self.language)}")

        return f'#EXT-X-SESSION-DATA:{",".join(session_data_out)}'

    def __str__(self) -> str:
        return self.dumps()


class DateRangeList(TagList):
    pass


class DateRange:
    def __init__(self, **kwargs: Any) -> None:
        self.id = kwargs["id"]
        self.start_date = kwargs.get("start_date")
        self.class_ = kwargs.get("class")
        self.end_date = kwargs.get("end_date")
        self.duration = kwargs.get("duration")
        self.planned_duration = kwargs.get("planned_duration")
        self.scte35_cmd = kwargs.get("scte35_cmd")
        self.scte35_out = kwargs.get("scte35_out")
        self.scte35_in = kwargs.get("scte35_in")
        self.end_on_next = kwargs.get("end_on_next")
        self.x_client_attrs = [(attr, kwargs.get(attr)) for attr in kwargs if attr.startswith("x_")]

    def dumps(self) -> str:
        daterange = [f"ID={quoted(self.id)}"]

        # whilst START-DATE is technically REQUIRED by the spec, this is
        # contradicted by an example in the same document (see
        # https://tools.ietf.org/html/rfc8216#section-8.10), and also by
        # real-world implementations, so we make it optional here
        if self.start_date:
            daterange.append(f"START-DATE={quoted(self.start_date)}")
        if self.class_:
            daterange.append(f"CLASS={quoted(self.class_)}")
        if self.end_date:
            daterange.append(f"END-DATE={quoted(self.end_date)}")
        if self.duration:
            daterange.append(f"DURATION={number_to_string(self.duration)}")
        if self.planned_duration:
            daterange.append(f"PLANNED-DURATION={number_to_string(self.planned_duration)}")
        if self.scte35_cmd:
            daterange.append(f"SCTE35-CMD={self.scte35_cmd}")
        if self.scte35_out:
            daterange.append(f"SCTE35-OUT={self.scte35_out}")
        if self.scte35_in:
            daterange.append(f"SCTE35-IN={self.scte35_in}")
        if self.end_on_next:
            daterange.append(f"END-ON-NEXT={self.end_on_next}")

        # client attributes sorted alphabetically output order is predictable
        daterange.extend(f"{denormalize_attribute(attr)}={value}" for attr, value in sorted(self.x_client_attrs))

        return f'#EXT-X-DATERANGE:{",".join(daterange)}'

    def __str__(self) -> str:
        return self.dumps()


class ContentSteering(BasePathMixin):
    def __init__(self, base_uri, server_uri, pathway_id=None) -> None:
        self.base_uri = base_uri
        self.uri = server_uri
        self.pathway_id = pathway_id

    def dumps(self) -> str:
        steering = [f"SERVER-URI={quoted(self.uri)}"]

        if self.pathway_id is not None:
            steering.append(f"PATHWAY-ID={quoted(self.pathway_id)}")

        return f'#EXT-X-CONTENT-STEERING:{",".join(steering)}'

    def __str__(self) -> str:
        return self.dumps()


def find_key(keydata, keylist: list[Key]) -> Key | None:
    if not keydata:
        return None
    for key in keylist:
        if key and (
            keydata.get("uri", None) == key.uri
            and keydata.get("method", "NONE") == key.method
            and keydata.get("iv", None) == key.iv
        ):
            return key
    raise KeyError("No key found for key data")


def denormalize_attribute(attribute: str) -> str:
    return attribute.replace("_", "-").upper()


def quoted(string: str) -> str:
    return f'"{string}"'


def number_to_string(number: int | float) -> str:
    with decimal.localcontext() as ctx:
        ctx.prec = 20  # set floating point precision
        d = decimal.Decimal(str(number))
        return f"{d.quantize(decimal.Decimal(1)) if d == d.to_integral_value() else d.normalize()}"



from __future__ import annotations

import hashlib
import struct
import typing
import uuid
from functools import total_ordering
from typing import Any

import discord
from cached_property import cached_property  # type: ignore
from cashews import Cache  # type: ignore

from pylav.constants.regex import SQUARE_BRACKETS, STREAM_TITLE
from pylav.exceptions.track import InvalidTrackException, TrackNotFoundException
from pylav.extension.flowery.lyrics import Error, Lyrics
from pylav.helpers.misc import MISSING
from pylav.nodes.api.responses.playlists import Info
from pylav.nodes.api.responses.track import Track as APITrack
from pylav.players.query.obj import Query
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

if typing.TYPE_CHECKING:
    from pylav.core.client import Client
    from pylav.nodes.node import Node
    from pylav.players.player import Player


__CLIENT: Client | None = None


# noinspection SpellCheckingInspection
@total_ordering
class Track:
    __slots__ = (
        "_node",
        "_is_partial",
        "_query",
        "_extra",
        "_unique_id",
        "__clear_cache_task",
        "_skip_segments",
        "_requester_id",
        "_requester",
        "_updated_query",
        "_id",
        "_extra",
        "_encoded",
        "_raw_data",
        "_processed",
    )
    __CLIENT: Client | None = None

    def __init__(
        self,
        node: Node,
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> None:
        """This class should not be instantiated directly. Use :meth:`Track.build_track` instead."""
        self._encoded: str | None = None
        self._is_partial = False
        self._requester = requester
        self._node = node
        self._query = query
        self._extra = extra
        self._skip_segments = skip_segments or []
        self._unique_id = hashlib.md5()
        self._updated_query = None
        self._id = str(uuid.uuid4())
        self._raw_data: JSON_DICT_TYPE = {}
        self._processed: APITrack | None = None

        self._process_init()

    @property
    def client(self) -> Client:
        """Get the client"""
        global __CLIENT
        return self.__CLIENT or __CLIENT

    @classmethod
    def attach_client(cls, client: Client) -> None:
        global __CLIENT
        __CLIENT = cls.__CLIENT = client

    def __int__(self) -> int:
        return 0

    def __gt__(self, other: Any) -> bool:
        return False

    def __lt__(self, other: Any) -> bool:
        return False

    def __ge__(self, other: Any) -> bool:
        return True

    def __le__(self, other: Any) -> bool:
        return True

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Track):
            return self._id == other._id
        raise NotImplemented

    def __ne__(self, other: Any) -> bool:
        x = self.__eq__(other)
        return not x

    def __hash__(self):
        return hash((self.unique_identifier,))

    def __getitem__(self, name: str) -> Any:
        return super().__getattribute__(name)

    def __repr__(self) -> str:
        return f"<Track identifier={self._id} encoded={self.encoded}>"

    def _process_init(self) -> None:
        if self._requester is None:
            self._requester_id = self.client.bot.user.id
        elif isinstance(self._requester, int):
            self._requester_id = self._requester
        else:
            self._requester_id = self._requester.id

        if "partial" in self._extra:
            self._is_partial = self._extra["partial"]

        if self._query is not None:
            self.timestamp = self.timestamp or self._query.start_time

        self._extra.pop("track", None)

    @classmethod
    async def build_track(
        cls,
        node: Node,
        data: Track | APITrack | dict[str, Any] | str | None,
        query: Query | None,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        lazy: bool = False,
        **extra: Any,
    ) -> Track:
        """Builds a track object from the given data.

        Parameters
        ----------
        node: None
            The node that the track is being built for.
        data: :class:`Track` | :class:`LavalinkTrackObject` | :class:`dict` | :class:`str` | None
            The data to build the track from.
        query: :class:`Query`
            The query that the track was built from.
        skip_segments: :class:`list`[:class:`str`]
            The segments to skip when playing the track.
        requester: :class:`discord.abc.User` | :class:`int`
            The user that requested the track.
        lazy: :class:`bool`
            Whether to build decode via the bot or Lavalink
        **extra: Any
            Extra data to add to the track.
        """

        # Check if data is a Track object and return it if it is
        if isinstance(data, cls):
            return data

        # Check if data is a LavalinkTrackObject and process it.
        if isinstance(data, APITrack):
            return cls._from_lavalink_track_object(node, data, query, skip_segments, requester, **extra)

        if isinstance(data, dict):
            try:
                return cls._from_mapping(node, data, query, skip_segments, requester, **extra)
            except KeyError as ke:
                (missing_key,) = ke.args
                raise InvalidTrackException(
                    f"Cannot build a track from partial data! (Missing key: {missing_key})"
                ) from ke

        if data is None:
            return cls._from_query(node, query, skip_segments, requester, **extra)

        if isinstance(data, str):
            return await cls._from_base64_string(node, data, skip_segments, requester, lazy=lazy, **extra)

        raise TypeError(f"Expected Track, LavalinkTrackObject, dict, or str, got {type(data).__name__}")

    @classmethod
    def _from_lavalink_track_object(
        cls,
        node: Node,
        data: APITrack,
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> Track:
        instance = cls(node, query, skip_segments, requester, **extra)
        instance._encoded = data.encoded
        instance._extra = extra
        instance._raw_data = extra.get("raw_data", {})
        instance._unique_id = hashlib.md5()
        instance._unique_id.update(instance._encoded.encode())
        instance._is_partial = False
        instance._processed = data
        return instance

    @classmethod
    def _from_mapping(
        cls,
        node: Node,
        data: dict[str, Any],
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> Track:
        instance = cls(node, query, skip_segments, requester, **extra)
        instance._encoded = data["encoded"]
        instance._is_partial = not bool(data["encoded"])
        instance._raw_data = data.get("raw_data", {}) or extra.get("raw_data", {})
        if instance._encoded is None:
            raise InvalidTrackException("Cannot build a track from partial data! (Missing key: encoded)")
        instance._unique_id = hashlib.md5()
        instance._unique_id.update(instance._encoded.encode())
        return instance

    @classmethod
    async def _from_base64_string(
        cls,
        node: Node,
        data: str,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        lazy: bool = False,
        **extra: Any,
    ) -> Track:
        track_obj = await cls.__CLIENT.decode_track(data, raise_on_failure=True, lazy=lazy)
        query_obj = await Query.from_string(track_obj.info.uri)
        return cls._from_lavalink_track_object(
            node=node, data=track_obj, query=query_obj, skip_segments=skip_segments, requester=requester, **extra
        )

    @classmethod
    def _from_query(
        cls,
        node: Node,
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> Track:
        instance = cls(node, query, skip_segments, requester, **extra)
        instance._extra = extra
        instance._raw_data = extra.get("raw_data", {})
        instance._is_partial = True
        return instance

    def _copy_with_extras(self, node: Node, **extra: Any) -> Track:
        instance = self.__class__(node, self._query, None, None)
        instance._extra = {**self._extra, **extra}
        instance._encoded = self._encoded
        instance._raw_data = self._raw_data
        instance._unique_id = self._unique_id
        instance._is_partial = self._is_partial
        instance._processed = self._processed
        return instance

    @property
    def skip_segments(self) -> list[str]:
        """The segments to skip when playing the track."""
        return self._skip_segments

    @property
    def encoded(self) -> str | None:
        return self._encoded

    @property
    def timestamp(self) -> int:
        return self._extra.get("timestamp", 0)

    @timestamp.setter
    def timestamp(self, value: int) -> None:
        self._extra["timestamp"] = value

    @property
    def requester_id(self) -> int:
        return self._requester_id

    @property
    def requester(self) -> discord.User | None:
        return self.client.bot.get_user(self.requester_id)

    @property
    def last_known_position(self) -> int:
        return self._extra.get("last_known_position", 0)

    @last_known_position.setter
    def last_known_position(self, value: int) -> None:
        self._extra["last_known_position"] = value

    @property
    def id(self) -> str:
        return self._id

    @property
    def unique_identifier(self) -> str:
        return self._unique_id.hexdigest()

    @property
    def is_partial(self) -> bool:
        return bool(self._is_partial and not self.encoded)

    async def identifier(self) -> str | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.identifier

    async def is_seekable(self) -> bool:
        return False if self.is_partial else (await self.fetch_full_track_data()).info.isSeekable

    async def duration(self) -> int:
        return 0 if self.is_partial else (await self.fetch_full_track_data()).info.length

    length = duration

    async def stream(self) -> bool:
        return False if self.is_partial else (await self.fetch_full_track_data()).info.isStream

    async def title(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.title

    async def uri(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.uri

    async def author(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.author

    async def source(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.sourceName

    async def artworkUrl(self) -> str | None:  # noqa:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.artworkUrl

    async def isrc(self) -> str | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.isrc

    async def info(self) -> Info | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info

    async def probe_info(self) -> str | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).pluginInfo.probeInfo

    async def query(self) -> Query:
        if self._processed and self._updated_query is None:
            self._updated_query = await Query.from_string(self._processed.info.uri)
            if self._query:
                self._updated_query.merge(
                    query=self._query,
                    start_time=True,
                    index=True,
                    source=True,
                    recursive=True,
                    search=True,
                )
            self._query = self._updated_query
            self.timestamp = self.timestamp or self._query.start_time
        if self.encoded and self._updated_query is None:
            assert self.encoded is not None
            self._updated_query = await Query.from_base64(self.encoded, lazy=True)
            if self._query:
                self._updated_query.merge(
                    query=self._query,
                    start_time=True,
                    index=True,
                    source=True,
                    recursive=True,
                    search=True,
                )
            self._query = self._updated_query
            self.timestamp = self.timestamp or self._query.start_time
        return self._query

    async def is_clypit(self) -> bool:
        return (await self.query()).is_clypit

    async def is_getyarn(self) -> bool:
        return (await self.query()).is_getyarn

    async def is_mixcloud(self) -> bool:
        return (await self.query()).is_mixcloud

    async def is_ocremix(self) -> bool:
        return (await self.query()).is_ocremix

    async def is_pornhub(self) -> bool:
        return (await self.query()).is_pornhub

    async def is_reddit(self) -> bool:
        return (await self.query()).is_reddit

    async def is_soundgasm(self) -> bool:
        return (await self.query()).is_soundgasm

    async def is_tiktok(self) -> bool:
        return (await self.query()).is_tiktok

    async def is_spotify(self) -> bool:
        return (await self.query()).is_spotify

    async def is_apple_music(self) -> bool:
        return (await self.query()).is_apple_music

    async def is_bandcamp(self) -> bool:
        return (await self.query()).is_bandcamp

    async def is_youtube(self) -> bool:
        return (await self.query()).is_youtube

    async def is_youtube_music(self) -> bool:
        return (await self.query()).is_youtube_music

    async def is_soundcloud(self) -> bool:
        return (await self.query()).is_soundcloud

    async def is_twitch(self) -> bool:
        return (await self.query()).is_twitch

    async def is_http(self) -> bool:
        return (await self.query()).is_http

    async def is_local(self) -> bool:
        return (await self.query()).is_local

    async def is_niconico(self) -> bool:
        return (await self.query()).is_niconico

    async def is_vimeo(self) -> bool:
        return (await self.query()).is_vimeo

    async def is_search(self) -> bool:
        return (await self.query()).is_search

    async def is_album(self) -> bool:
        return (await self.query()).is_album

    async def is_playlist(self) -> bool:
        return (await self.query()).is_playlist

    async def is_single(self) -> bool:
        return (await self.query()).is_single

    async def is_speak(self) -> bool:
        return (await self.query()).is_speak

    async def is_gctts(self) -> bool:
        return (await self.query()).is_gctts

    async def is_deezer(self) -> bool:
        return (await self.query()).is_deezer

    async def requires_capability(self) -> str:
        return q.requires_capability if (q := await self.query()) else "youtube"

    async def query_identifier(self) -> str:
        return (await self.query()).query_identifier

    async def query_source(self) -> str:
        return (await self.query()).source

    async def mix_playlist_url(self) -> str | None:
        if not await self.identifier():
            return None
        if await self.source() == "youtube":
            return f"https://www.youtube.com/watch?v={self.identifier}&list=RD{self.identifier}"
        return None

    async def fetch_full_track_data(
        self,
    ) -> APITrack:
        if self._processed:
            return self._processed
        if self.encoded:
            self._processed = await self.client.decode_track(self.encoded)
        else:
            await self.search()
        return self._processed

    async def to_dict(self) -> dict[str, Any]:
        """
        Returns a dict representation of this Track.
        Returns
        -------
        :class:`dict`
            The dict representation of this Track.
        """
        return {
            "encoded": self.encoded,
            "query": await self.query_identifier(),
            "requester": self.requester.id if self.requester else self.requester_id,
            "skip_segments": self._skip_segments,
            "extra": {
                "timestamp": self.timestamp,
                "last_known_position": self.last_known_position,
                "partial": self._is_partial,
            },
            "raw_data": self._raw_data,
        }

    async def search(self, bypass_cache: bool = False):
        _query = await Query.from_string(self._query)
        _query.merge(
            query=self._query,
            start_time=True,
            index=True,
            source=True,
            recursive=True,
            search=True,
        )
        self._query = _query
        self.timestamp = self.timestamp or self._query.start_time
        response = await self.client._get_tracks(await self.query(), first=True, bypass_cache=bypass_cache)
        if not response or not response.tracks:
            raise TrackNotFoundException(f"No tracks found for query {await self.query_identifier()}")
        track = response.tracks[0]
        self._encoded = track.encoded
        assert isinstance(self._encoded, str)
        self._unique_id = hashlib.md5()
        self._unique_id.update(self.encoded.encode())
        self._is_partial = False
        self._processed = track

    async def search_all(self, player: Player, requester: int, bypass_cache: bool = False) -> list[Track]:
        _query = await Query.from_string(self._query)
        _query.merge(
            query=self._query,
            start_time=True,
            index=True,
            source=True,
            recursive=True,
            search=True,
        )
        self._query = _query
        response = await player.node.get_track(
            await self.query(), bypass_cache=bypass_cache, first=self._query.is_search
        )
        if not response or not response.tracks:
            raise TrackNotFoundException(f"No tracks found for query {await self.query_identifier()}")
        return [
            await Track.build_track(data=track, node=player.node, query=self._query, requester=requester)
            for track in response.tracks
        ]

    async def _icyparser(self, url: str) -> str | None:
        try:
            async with self._node.session.get(url, headers={"Icy-MetaData": "1"}) as resp:
                metaint = int(resp.headers["icy-metaint"])
                for __ in iter(range(5)):
                    await resp.content.readexactly(metaint)
                    metadata_length = struct.unpack("B", await resp.content.readexactly(1))[0] * 16
                    metadata = await resp.content.readexactly(metadata_length)
                    if not (m := STREAM_TITLE.search(metadata.rstrip(b"\0"))):
                        return None
                    if title := m.group(1):
                        title = title.decode("utf-8", errors="replace")
                        return title
        except Exception:  # noqa
            return None
        return None

    async def get_track_display_name(
        self,
        max_length: int | None = None,
        author: bool = True,
        unformatted: bool = False,
        with_url: bool = False,
        escape: bool = True,
    ) -> str:
        if unformatted:
            return await self.get_track_display_name_unformatted(max_length=max_length, author=author, escape=escape)
        else:
            return await self.get_track_display_name_formatted(
                max_length=max_length, author=author, with_url=with_url, escape=escape
            )

    async def get_track_display_name_unformatted(
        self,
        max_length: int | None = None,
        author: bool = True,
        escape: bool = True,
    ) -> str:
        if self.is_partial:
            track_name = await self.get_partial_track_display_name(max_length=max_length)
        else:
            track_name = await self.get_full_track_display_name(max_length=max_length, author=author)
        return self._maybe_escape_markdown(text=track_name, escape=escape)

    async def get_track_display_name_formatted(
        self,
        max_length: int | None = None,
        author: bool = True,
        with_url: bool = False,
        escape: bool = True,
    ) -> str:
        if self.is_partial:
            track_name = await self.get_partial_track_display_name(
                max_length=max_length if with_url and max_length is None else (max_length - 8)
            )
            track_name = self._maybe_escape_markdown(text=track_name, escape=escape)
            if with_url and (query := await self.query()):
                if not query.is_single and not query.is_custom_playlist and not query.is_local:
                    track_name = f"**[{track_name}]({query.query_identifier})**"
        else:
            track_name = await self.get_full_track_display_name(
                max_length=max_length if with_url and max_length is None else (max_length - 8), author=author
            )
            track_name = self._maybe_escape_markdown(text=track_name, escape=escape)
            if with_url and (query := await self.query()) and not query.is_local:
                track_name = f"**[{track_name}]({await self.uri()})**"

        return track_name

    async def get_partial_track_display_name(self, max_length: int | None = None) -> str:
        query = await self.query()
        track_name = await query.query_to_queue(max_length if max_length is None else (max_length - 1), partial=True)
        track_name = SQUARE_BRACKETS.sub("", track_name).strip()
        if max_length is not None and len(track_name) > (max_length - 1):
            max_length -= 1
            return f"{track_name[:max_length]}\N{HORIZONTAL ELLIPSIS}"
        return track_name

    async def get_full_track_display_name(self, max_length: int | None = None, author: bool = True) -> str:
        author_string = f" - {await self.author()}" if author else ""
        return (
            await self.get_local_query_track_display_name(
                max_length=max_length,
                author_string=author_string,
                unknown_author=await self.author() != "Unknown artist",
                unknown_title=await self.title() != "Unknown title",
            )
            if await self.query() and await self.is_local()
            else await self.get_external_query_track_display_name(max_length=max_length, author_string=author_string)
        )

    async def get_local_query_track_display_name(
        self,
        author_string: str,
        unknown_author: bool,
        unknown_title: bool,
        max_length: int | None = None,
    ) -> str:
        if not (unknown_title and unknown_author):
            track_name = f"{await self.title()}{author_string}"
            track_name = SQUARE_BRACKETS.sub("", track_name).strip()
            if max_length and len(track_name) > (max_length - 1):
                max_length -= 1
                track_name = f"{track_name[:max_length]}\N{HORIZONTAL ELLIPSIS}"
            elif not max_length:
                track_name += f"\n{await (await self.query()).query_to_string(add_ellipsis=False, no_extension=True)} "
        else:
            track_name = await (await self.query()).query_to_string(max_length, name_only=True, no_extension=True)
            track_name = SQUARE_BRACKETS.sub("", track_name).strip()
        return track_name

    async def get_external_query_track_display_name(self, author_string: str, max_length: int | None = None) -> str:
        title = await self.title()

        if await self.stream():
            icy = await self._icyparser(await self.uri())
            track_name = icy or f"{title}{author_string}"
        elif (await self.author()).lower() not in title.lower():
            track_name = f"{title}{author_string}"
        else:
            track_name = title

        track_name = SQUARE_BRACKETS.sub("", track_name).strip()
        if max_length is not None and len(track_name) > (max_length - 1):
            max_length -= 1
            return f"{track_name[:max_length]}\N{HORIZONTAL ELLIPSIS}"
        return track_name

    @staticmethod
    def _maybe_escape_markdown(text: str, escape: bool = True) -> str:
        return discord.utils.escape_markdown(text) if escape else text

    async def fetch_lyrics(self) -> tuple[bool, Lyrics | Error | None]:
        if isrc := await self.isrc():
            return True, await self.client.flowery_api.lyrics.get_lyrics(isrc=isrc)
        elif not self.is_partial:
            if await self.is_spotify():
                return True, await self.client.flowery_api.lyrics.get_lyrics(spotify_id=await self.identifier())
            elif await self.source() in {"deezer", "applemusic"}:
                return True, await self.client.flowery_api.lyrics.get_lyrics(
                    query=f"{await self.title()} artist:{await self.author()}"
                )
            else:
                return False, await self.client.flowery_api.lyrics.get_lyrics(query=await self.title())
        return False, None



from __future__ import annotations

import asyncio
from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.extension.red.ui.selectors.options.playlist import PlaylistOption
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.playlist import Playlist
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class PlaylistSelectSelector(discord.ui.Select):
    def __init__(
        self,
        options: list[PlaylistOption],
        cog: DISCORD_COG_TYPE,
        placeholder: str,
        mapping: dict[str, Playlist],
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            options=options,
            placeholder=shorten_string(max_length=100, string=placeholder),
        )
        self.cog = cog
        self.mapping = mapping
        self.playlist: Playlist | None = None
        self.responded = asyncio.Event()

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        playlist_id = self.values[0]
        self.playlist: Playlist = self.mapping.get(playlist_id)
        if self.playlist is None:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(messageable=interaction, title=_("Playlist was not found.")),
                ephemeral=True,
            )
            self.view.stop()
            await self.view.on_timeout()
            return
        self.responded.set()
        self.view.stop()
        await self.view.on_timeout()


class PlaylistPlaySelector(discord.ui.Select):
    def __init__(
        self,
        options: list[PlaylistOption],
        cog: DISCORD_COG_TYPE,
        placeholder: str,
        mapping: dict[str, Playlist],
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            options=options,
            placeholder=shorten_string(max_length=100, string=placeholder),
        )
        self.cog = cog
        self.mapping = mapping

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option.")
                ),
                ephemeral=True,
            )
            return
        playlist_id = self.values[0]
        playlist: Playlist = self.mapping.get(playlist_id)
        if playlist is None:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(messageable=interaction, title=_("Playlist was not found.")),
                ephemeral=True,
            )
            self.view.stop()
            await self.view.on_timeout()
            return
        await self.cog.command_playlist_play.callback(self.cog, interaction, playlist=[playlist])
        self.view.stop()
        await self.view.on_timeout()



from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Literal

import discord

from pylav.events.base import PyLavEvent
from pylav.nodes.node import Node

if TYPE_CHECKING:
    from pylav.players.filters import (
        ChannelMix,
        Distortion,
        Echo,
        Equalizer,
        Karaoke,
        LowPass,
        Rotation,
        Timescale,
        Tremolo,
        Vibrato,
        Volume,
    )
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class PlayerUpdateEvent(PyLavEvent):
    """This event is dispatched when the player's progress changes."""

    __slots__ = ()

    def __init__(self, player: Player, position: float, timestamp: float) -> None:
        self.player = player
        self.position = position
        self.timestamp = timestamp


class PlayerPausedEvent(PyLavEvent):
    """This event is dispatched when a player is paused."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class PlayerStoppedEvent(PyLavEvent):
    """This event is dispatched when a player is stopped."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class PlayerResumedEvent(PyLavEvent):
    """This event is dispatched when a player is resumed."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class PlayerRestoredEvent(PyLavEvent):
    """This event is dispatched when a player is restored."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.abc.User) -> None:
        self.player = player
        self.requester = requester


class PlayerMovedEvent(PyLavEvent):
    """This event is dispatched when the player is moved."""

    __slots__ = ()

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        before: discord.channel.VocalGuildChannel,
        after: discord.channel.VocalGuildChannel,
    ) -> None:
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after


class PlayerDisconnectedEvent(PyLavEvent):
    """This event is dispatched when the player is disconnected."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester
        self.current_track: Track | None = player.current
        self.position: float = player.estimated_position
        self.queue: collections.deque[Track] = player.queue.raw_queue


class PlayerConnectedEvent(PyLavEvent):
    """This event is dispatched when the player is connected."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member | None) -> None:
        self.player = player
        self.requester = requester


class PlayerVolumeChangedEvent(PyLavEvent):
    """This event is dispatched when the player has its volume changed."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member, before: int, after: int) -> None:
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after


class PlayerRepeatEvent(PyLavEvent):
    """This event is dispatched when the player repeat config changed."""

    __slots__ = ()

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        op_type: Literal["current", "queue"],
        queue_before: bool,
        queue_after: bool,
        current_before: bool,
        current_after: bool,
    ) -> None:
        self.player = player
        self.requester = requester
        self.queue_before = queue_before
        self.queue_after = queue_after
        self.current_before = current_before
        self.current_after = current_after
        self.type = op_type


class FiltersAppliedEvent(PyLavEvent):
    """This event is dispatched when filters are applied."""

    __slots__ = ()

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        node: Node,
        volume: Volume | None = None,
        equalizer: Equalizer | None = None,
        karaoke: Karaoke | None = None,
        timescale: Timescale | None = None,
        tremolo: Tremolo | None = None,
        vibrato: Vibrato | None = None,
        rotation: Rotation | None = None,
        distortion: Distortion | None = None,
        low_pass: LowPass | None = None,
        channel_mix: ChannelMix | None = None,
        echo: Echo | None = None,
    ) -> None:
        self.player = player
        self.requester = requester
        self.volume = volume
        self.equalizer = equalizer
        self.karaoke = karaoke
        self.timescale = timescale
        self.tremolo = tremolo
        self.vibrato = vibrato
        self.rotation = rotation
        self.distortion = distortion
        self.low_pass = low_pass
        self.channel_mix = channel_mix
        self.node = node
        self.echo = echo


class QuickPlayEvent(PyLavEvent):
    """This event is dispatched when a track is played with higher priority than the current track."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member, track: Track) -> None:
        self.player = player
        self.requester = requester
        self.track = track


class PlayerAutoDisconnectedEmptyQueueEvent(PyLavEvent):
    """This event is dispatched when the player is auto disconnected."""

    __slots__ = ()

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me
        self.current_track: Track | None = player.current
        self.position: float = player.estimated_position
        self.queue: collections.deque[Track] = player.queue.raw_queue


class PlayerAutoDisconnectedAloneEvent(PyLavEvent):
    """This event is dispatched when the player is auto disconnected."""

    __slots__ = ()

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me
        self.current_track: Track | None = player.current
        self.position: float = player.estimated_position
        self.queue: collections.deque[Track] = player.queue.raw_queue


class PlayerAutoPausedEvent(PyLavEvent):
    """This event is dispatched when the player is auto paused."""

    __slots__ = ()

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me


class PlayerAutoResumedEvent(PyLavEvent):
    """This event is dispatched when the player is auto resumed."""

    __slots__ = ()

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me
